"""
:author: David Siroky (siroky@dasir.cz)
:license: MIT License (see LICENSE.txt or
          U{http://www.opensource.org/licenses/mit-license.php})
"""

CERT_NONE = None
PROTOCOL_SSLv23 = None

class SSLError(StandardError):
    pass
